from django.conf.urls import url, include

from pretix.multidomain import event_url
from .views import auth_start, auth_disconnect, webhook, ReturnView, redirect_view

event_patterns = [
    url(r'^bitpay/', include([
        event_url(r'^webhook/$', webhook, name='webhook', require_live=False),
        event_url(r'^redirect/$', redirect_view, name='redirect', require_live=False),
        url(r'^return/(?P<order>[^/]+)/(?P<hash>[^/]+)/(?P<payment>[^/]+)/$', ReturnView.as_view(), name='return'),
    ])),
]
urlpatterns = [
    url(r'^control/event/(?P<organizer>[^/]+)/(?P<event>[^/]+)/bitpay/connect/',
        auth_start, name='auth.start'),
    url(r'^control/event/(?P<organizer>[^/]+)/(?P<event>[^/]+)/bitpay/disconnect/',
        auth_disconnect, name='auth.disconnect'),
]



import hashlib
import json
import logging
import urllib.parse
from decimal import Decimal

import requests
from django.conf import settings
from django.contrib import messages
from django.core import signing
from django.db import transaction
from django.http import Http404, HttpResponse, HttpResponseBadRequest
from django.shortcuts import redirect, render, get_object_or_404
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django.views import View
from django.views.decorators.clickjacking import xframe_options_exempt
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST

from pretix.base.models import Order, Quota, OrderPayment
from pretix.base.services.locking import LockTimeoutException
from pretix.control.permissions import event_permission_required
from pretix.multidomain.urlreverse import eventreverse, build_absolute_uri
from .models import ReferencedBitPayObject
from .payment import BitPay

logger = logging.getLogger(__name__)


@xframe_options_exempt
def redirect_view(request, *args, **kwargs):
    try:
        data = signing.loads(request.GET.get('data', ''), salt='safe-redirect')
    except signing.BadSignature:
        return HttpResponseBadRequest('Invalid parameter')

    if 'go' in request.GET:
        if 'session' in data:
            for k, v in data['session'].items():
                request.session[k] = v
        return redirect(data['url'])
    else:
        params = request.GET.copy()
        params['go'] = '1'
        r = render(request, 'pretix_bitpay/redirect.html', {
            'url': build_absolute_uri(request.event, 'plugins:pretix_bitpay:redirect') + '?' + urllib.parse.urlencode(params),
        })
        r._csp_ignore = True
        return r


@csrf_exempt
@require_POST
def webhook(request, *args, **kwargs):
    event_json = json.loads(request.body.decode('utf-8'))
    if 'data' not in event_json and 'id' in event_json:
        objid = event_json['id']
    else:
        objid = event_json['data']['id']
    try:
        rso = ReferencedBitPayObject.objects.select_related('order', 'order__event').get(reference=objid)
        if rso.order.event != request.event:
            return HttpResponse("Unable to detect event", status=200)
        rso.order.log_action('pretix_bitpay.event', data=event_json)
        return process_invoice(rso.order, rso.payment, objid)
    except ReferencedBitPayObject.DoesNotExist:
        return HttpResponse("Unable to detect event", status=200)


def process_invoice(order, payment, invoice_id):
    prov = BitPay(order.event)
    src = prov.client.get_invoice(invoice_id)

    if not payment:
        payment = order.payments.filter(
            info__icontains=invoice_id,
            provider='bitpay',
        ).last()
    if not payment:
        payment = order.payments.create(
            state=OrderPayment.PAYMENT_STATE_CREATED,
            provider='bitpay',
            amount=Decimal(src['amount']),
            info=json.dumps(src),
        )

    with transaction.atomic():
        order.refresh_from_db()
        payment.refresh_from_db()

        if src['status'] == 'new':
            pass
        elif src['status'] in ('paid', 'confirmed', 'complete'):
            if payment.state in (OrderPayment.PAYMENT_STATE_CREATED, OrderPayment.PAYMENT_STATE_PENDING):
                try:
                    payment.confirm()
                except LockTimeoutException:
                    return HttpResponse("Lock timeout, please try again.", status=503)
                except Quota.QuotaExceededException:
                    return HttpResponse("Quota exceeded.", status=200)
        elif src['status'] == ('expired', 'invalid'):
            if payment.state in (OrderPayment.PAYMENT_STATE_CREATED, OrderPayment.PAYMENT_STATE_PENDING):
                payment.state = OrderPayment.PAYMENT_STATE_FAILED
                payment.info = json.dumps(src)
                payment.save()
                payment.order.log_action('pretix.event.order.payment.failed', {
                    'local_id': payment.local_id,
                    'provider': payment.provider,
                    'info': json.dumps(src)
                })
            elif order.state in OrderPayment.PAYMENT_STATE_CONFIRMED:
                payment.state = OrderPayment.PAYMENT_STATE_FAILED
                payment.info = json.dumps(src)
                payment.save()
                payment.create_external_refund()
                payment.order.log_action('pretix.event.order.payment.failed', {
                    'local_id': payment.local_id,
                    'provider': payment.provider,
                    'info': json.dumps(src)
                })

    return HttpResponse(status=200)


@event_permission_required('can_change_event_settings')
@require_POST
def auth_disconnect(request, **kwargs):
    del request.event.settings.payment_bitpay_token
    request.event.settings.payment_bitpay__enabled = False
    messages.success(request, _('Your BitPay account has been disconnected.'))

    return redirect(reverse('control:event.settings.payment.provider', kwargs={
        'organizer': request.event.organizer.slug,
        'event': request.event.slug,
        'provider': 'bitpay'
    }))


class BitPayOrderView:
    def dispatch(self, request, *args, **kwargs):
        try:
            self.order = request.event.orders.get(code=kwargs['order'])
            if hashlib.sha1(self.order.secret.lower().encode()).hexdigest() != kwargs['hash'].lower():
                raise Http404('')
        except Order.DoesNotExist:
            # Do a hash comparison as well to harden timing attacks
            if 'abcdefghijklmnopq'.lower() == hashlib.sha1('abcdefghijklmnopq'.encode()).hexdigest():
                raise Http404('')
            else:
                raise Http404('')
        return super().dispatch(request, *args, **kwargs)

    @cached_property
    def pprov(self):
        return self.payment.payment_provider

    @property
    def payment(self):
        return get_object_or_404(
            self.order.payments,
            pk=self.kwargs['payment'],
            provider__istartswith='bitpay',
        )


@method_decorator(xframe_options_exempt, 'dispatch')
class ReturnView(BitPayOrderView, View):
    def get(self, request, *args, **kwargs):
        with transaction.atomic():
            self.order.refresh_from_db()
            if self.order.status == Order.STATUS_PAID:
                return self._redirect_to_order()

            payment_data = self.payment.info_data
            process_invoice(self.order, self.payment, payment_data['id'])
            return self._redirect_to_order()

    def _redirect_to_order(self):
        if self.request.session.get('payment_bitpay_order_secret') != self.order.secret:
            messages.error(self.request, _('Sorry, there was an error in the payment process. Please check the link '
                                           'in your emails to continue.'))
            return redirect(eventreverse(self.request.event, 'presale:event.index'))

        return redirect(eventreverse(self.request.event, 'presale:event.order', kwargs={
            'order': self.order.code,
            'secret': self.order.secret
        }) + ('?paid=yes' if self.order.status == Order.STATUS_PAID else ''))


@event_permission_required('can_change_event_settings')
def auth_start(request, **kwargs):
    from btcpay import crypto  # improve performance at import time

    if request.event.settings.payment_bitpay_token:
        messages.error(request, _('You are already connected to BitPay.'))
        return redirect(reverse('control:event.settings.payment.provider', kwargs={
            'organizer': request.event.organizer.slug,
            'event': request.event.slug,
            'provider': 'bitpay'
        }))
    request.session['payment_bitpay_auth_event'] = request.event.pk

    request.event.settings.payment_bitpay_pem = crypto.generate_privkey()

    pem = request.event.settings.payment_bitpay_pem
    sin = crypto.get_sin_from_pem(pem)
    if request.GET.get('url'):
        url = request.GET.get('url')
    else:
        url = 'https://test.bitpay.com' if 'test' in request.GET else 'https://bitpay.com'
    try:
        r = requests.post(
            url + '/tokens',
            json={
                'label': settings.PRETIX_INSTANCE_NAME,
                'facade': 'merchant',
                'id': sin
            }
        )
    except requests.ConnectionError:
        messages.error(request, _('Communication with BitPay was not successful.'))
    else:
        if r.status_code == 200:
            data = r.json()['data'][0]
            request.event.settings.payment_bitpay_token = data['token']
            request.event.settings.payment_bitpay_url = url
            return redirect(
                url + '/api-access-request?pairingCode=' + data['pairingCode']
            )
        messages.error(request, _('Communication with BitPay was not successful.'))

    return redirect(reverse('control:event.settings.payment.provider', kwargs={
        'organizer': request.event.organizer.slug,
        'event': request.event.slug,
        'provider': 'bitpay'
    }))



import json

from django.dispatch import receiver
from django.template.loader import get_template
from django.utils.translation import gettext_lazy as _

from pretix.base.signals import (
    logentry_display,
    register_payment_providers, requiredaction_display, )


@receiver(register_payment_providers, dispatch_uid="payment_bitpay")
def register_payment_provider(sender, **kwargs):
    from .payment import BitPay
    return BitPay


@receiver(signal=logentry_display, dispatch_uid="bitpay_logentry_display")
def pretixcontrol_logentry_display(sender, logentry, **kwargs):
    if logentry.action_type != 'pretix_bitpay.event':
        return

    data = json.loads(logentry.data)
    event_type = data.get('event', {}).get('name', '?')
    return _('BitPay reported an event: {}').format(event_type)


@receiver(signal=requiredaction_display, dispatch_uid="bitpay_requiredaction_display")
def pretixcontrol_action_display(sender, action, request, **kwargs):
    if not action.action_type.startswith('pretix_bitpay'):
        return

    data = json.loads(action.data)

    if action.action_type == 'pretix_bitpay.refund':
        template = get_template('pretix_bitpay/action_refund.html')
    elif action.action_type == 'pretix_bitpay.overpaid':
        template = get_template('pretix_bitpay/action_overpaid.html')

    ctx = {'data': data, 'event': sender, 'action': action}
    return template.render(ctx, request)



import os
from distutils.command.build import build

from django.core import management
from setuptools import setup, find_packages

from pretix_bitpay import __version__

try:
    with open(os.path.join(os.path.dirname(__file__), 'README.rst'), encoding='utf-8') as f:
        long_description = f.read()
except:
    long_description = ''


class CustomBuild(build):
    def run(self):
        management.call_command('compilemessages', verbosity=1)
        build.run(self)


cmdclass = {
    'build': CustomBuild
}


setup(
    name='pretix-bitpay',
    version=__version__,
    description='This plugin allows accepting crypto currency payments in pretix via BitPay-compatible payment '
                'providers.',
    long_description=long_description,
    url='https://github.com/pretix/pretix-bitpay',
    author='Raphael Michel',
    author_email='michel@rami.io',
    license='Apache Software License',

    install_requires=['btcpay-python'],
    packages=find_packages(exclude=['tests', 'tests.*']),
    include_package_data=True,
    cmdclass=cmdclass,
    entry_points="""
[pretix.plugin]
pretix_bitpay=pretix_bitpay:PretixPluginMeta
""",
)



from typing import Union

import hashlib
import json
import logging
import requests
import urllib
from collections import OrderedDict
from django import forms
from django.core import signing
from django.http import HttpRequest
from django.template.loader import get_template
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _  # NoQA
from i18nfield.forms import I18nFormField, I18nTextInput
from i18nfield.strings import LazyI18nString
from json import JSONDecodeError
from pretix.base.models import OrderPayment, OrderRefund
from pretix.base.payment import BasePaymentProvider, PaymentException
from pretix.multidomain.urlreverse import build_absolute_uri
from requests import RequestException

from .models import ReferencedBitPayObject

logger = logging.getLogger(__name__)


class BitPay(BasePaymentProvider):
    identifier = 'bitpay'
    verbose_name = _('BitPay')

    @property
    def public_name(self):
        return str(self.settings.get('public_name', as_type=LazyI18nString) or _('Crypto'))

    def settings_content_render(self, request):
        if not self.settings.token:
            return (
                "<p>{}</p>"
                "<p>"
                "<a href='{}' target='_blank' class='btn btn-primary btn-lg'>{}</a> "
                "<a href='{}' target='_blank' class='btn btn-default btn-lg'>{}</a> "
                "</p>"
                "<p>{}</p>"
                "</form>"  # This is a hell of a hack, sorry…
                "<form class='form-inline' action='{}' method='GET' target='_blank'>"
                "<input type='text' name='url' class='form-control' value='https://btcpay.lightbo.lt'> "
                "<button class='btn btn-default'>{}</button>"
                "</form>"
                "<form>"
            ).format(
                _('To accept payments via BitPay, you will need an account at BitPay. By clicking on the '
                  'following button, you can connect pretix to your BitPay account. A BitPay site will open in a new '
                  'tab. Return to this page and refresh it after you authorized the token at BitPay.'),
                reverse('plugins:pretix_bitpay:auth.start', kwargs={
                    'organizer': self.event.organizer.slug,
                    'event': self.event.slug,
                }),
                _('Connect with BitPay'),
                reverse('plugins:pretix_bitpay:auth.start', kwargs={
                    'organizer': self.event.organizer.slug,
                    'event': self.event.slug,
                }) + '?test=1',
                _('Connect with test.bitpay.com'),
                _('Alternatively, you can connect with a third-party provider using a BitPay-compatible API. Enter '
                  'the URL here you want to connect to.'),
                reverse('plugins:pretix_bitpay:auth.start', kwargs={
                    'organizer': self.event.organizer.slug,
                    'event': self.event.slug,
                }),
                _('Start pairing')
            )
        else:
            return (
                "<button formaction='{}' class='btn btn-danger'>{}</button>"
            ).format(
                reverse('plugins:pretix_bitpay:auth.disconnect', kwargs={
                    'organizer': self.event.organizer.slug,
                    'event': self.event.slug,
                }),
                _('Disconnect from BitPay')
            )

    @property
    def settings_form_fields(self):
        if not self.settings.token:
            return {}
        d = OrderedDict(
            [
                ('url',
                 forms.CharField(
                     label=_('API URL'),
                     disabled=True
                 )),
                ('public_name', I18nFormField(
                    label=_('Payment method name'),
                    widget=I18nTextInput,
                    help_text=_(
                        'Since you can accept a variety of different Crypto coins with BitPay and BitPay-compatible '
                        'services, you can set the name of the payment method here to reflect the coins you are '
                        'actually accepting.'
                    )
                )),
            ] + list(super().settings_form_fields.items())
        )

        d.move_to_end('public_name', last=False)
        d.move_to_end('_enabled', last=False)
        return d

    def redirect(self, request, url):
        if request.session.get('iframe_session', False):
            return (
                build_absolute_uri(request.event, 'plugins:pretix_bitpay:redirect') +
                '?data=' + signing.dumps({
                    'url': url,
                    'session': {
                        'payment_bitpay_order_secret': request.session['payment_bitpay_order_secret'],
                    },
                }, salt='safe-redirect')
            )
        else:
            return str(url)

    def payment_form_render(self, request) -> str:
        template = get_template('pretix_bitpay/checkout_payment_form.html')
        ctx = {'request': request, 'event': self.event, 'settings': self.settings}
        return template.render(ctx)

    def checkout_confirm_render(self, request) -> str:
        template = get_template('pretix_bitpay/checkout_payment_confirm.html')
        ctx = {'request': request, 'event': self.event, 'settings': self.settings}
        return template.render(ctx)

    def checkout_prepare(self, request, total):
        return True

    def payment_is_valid_session(self, request):
        return True

    @property
    def test_mode_message(self):
        if '//test.bitpay.com' in self.settings.url:
            return _('The BitPay plugin is operating in test mode. No money will actually be transferred.')
        return None

    @cached_property
    def client(self):
        from btcpay import BTCPayClient
        return BTCPayClient(host=self.settings.url, pem=self.settings.pem, tokens={'merchant': self.settings.token})

    def execute_payment(self, request: HttpRequest, payment: OrderPayment):
        request.session['payment_bitpay_order_secret'] = payment.order.secret

        try:
            inv = self.client.create_invoice({
                "price": float(payment.amount),
                "currency": self.event.currency,
                "orderId": payment.order.full_code,
                "transactionSpeed": "medium",
                "extendedNotifications": "true",
                "notificationURL": build_absolute_uri(self.event, "plugins:pretix_bitpay:webhook"),
                "redirectURL": build_absolute_uri(self.event, "plugins:pretix_bitpay:return", kwargs={
                    'order': payment.order.code,
                    'payment': payment.pk,
                    'hash': hashlib.sha1(payment.order.secret.lower().encode()).hexdigest(),
                }),
                # "buyer": {"email": "test@customer.com"},
                "token": self.settings.token
            })
        except RequestException:
            logger.exception('Failure during bitpay payment.')
            raise PaymentException(_('We had trouble communicating with BitPay. Please try again and get in touch '
                                     'with us if this problem persists.'))
        ReferencedBitPayObject.objects.get_or_create(order=payment.order, payment=payment, reference=inv['id'])
        payment.info = json.dumps(inv)
        payment.save(update_fields=['info'])
        return self.redirect(request, inv['url'])

    def payment_pending_render(self, request: HttpRequest, payment: OrderPayment):
        template = get_template('pretix_bitpay/pending.html')
        ctx = {'request': request, 'event': self.event, 'settings': self.settings,
               'order': payment.order}
        return template.render(ctx)

    def payment_control_render(self, request: HttpRequest, payment: OrderPayment):
        template = get_template('pretix_bitpay/control.html')
        ctx = {'request': request, 'event': self.event, 'settings': self.settings,
               'payment_info': payment.info_data, 'order': payment.order, 'provname': self.verbose_name}
        return template.render(ctx)

    abort_pending_allowed = True

    def payment_refund_supported(self, payment: OrderPayment):
        return True

    def payment_partial_refund_supported(self, payment: OrderPayment):
        return True

    def _refund(self, refund):
        from btcpay import crypto
        payload = json.dumps({
            'token': refund.payment.info_data.get('token'),
            'amount': refund.payment.info_data.get('price'),
            'currency': refund.payment.info_data.get('currency'),
            'refundEmail': refund.order.email
        })
        uri = self.client.host + "/invoices/" + refund.payment.info_data.get('id') + "/refunds"
        xidentity = crypto.get_compressed_public_key_from_pem(self.client.pem)
        xsignature = crypto.sign(uri + payload, self.client.pem)
        headers = {"content-type": "application/json", 'accept': 'application/json', 'X-Identity': xidentity,
                   'X-Signature': xsignature, 'X-accept-version': '2.0.0'}
        try:
            response = requests.post(uri, data=payload, headers=headers, verify=self.client.verify)
        except Exception:
            logger.exception('Failure during bitpay refund.')
            raise PaymentException(_('We had trouble communicating with BitPay. Please try again and get in touch '
                                     'with us if this problem persists.'))
        if not response.ok:
            try:
                e = response.json()['error']
            except JSONDecodeError:
                e = response

            logger.exception('Failure during bitpay refund: {}'.format(e))
            raise PaymentException(_('BitPay reported an error: {}').format(e))

    def execute_refund(self, refund: OrderRefund):
        try:
            self._refund(refund)
        except requests.exceptions.RequestException as e:
            logger.exception('BitPay error: %s' % str(e))
            raise PaymentException(_('We had trouble communicating with BitPay. Please try again and contact '
                                     'support if the problem persists.'))
        else:
            refund.done()

    def shred_payment_info(self, obj: Union[OrderPayment, OrderRefund]):
        d = obj.info_data
        for k, v in list(d.items()):
            if v not in ('id', 'status', 'price', 'currency', 'invoiceTime', 'paymentSubtotals',
                         'paymentTotals', 'transactionCurrency', 'amountPaid'):
                d[k] = '█'
        obj.info_data = d
        obj.save()

        for le in obj.order.all_logentries().filter(action_type="pretix_bitpay.event").exclude(data=""):
            d = le.parsed_data
            if 'data' in d:
                for k, v in list(d['data'].items()):
                    if v not in ('id', 'status', 'price', 'currency', 'invoiceTime', 'paymentSubtotals',
                                 'paymentTotals', 'transactionCurrency', 'amountPaid'):
                        d['data'][k] = '█'
                le.data = json.dumps(d)
                le.shredded = True
                le.save(update_fields=['data', 'shredded'])



from django.apps import AppConfig
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

__version__ = "1.4.4"


class BitpayApp(AppConfig):
    name = 'pretix_bitpay'
    verbose_name = _("BitPay")

    class PretixPluginMeta:
        name = _("BitPay")
        author = "Raphael Michel"
        category = 'PAYMENT'
        version = __version__
        description = _("Receive Crypto payments via BitPay-compatible payment providers.")
        picture = "pretix_bitpay/bitpay-logo.svg"

    def ready(self):
        from . import signals   # NOQA

    @cached_property
    def compatibility_errors(self):
        errs = []
        try:
            import btcpay  # NOQA
        except ImportError:
            errs.append("Python package 'btcpay' is not installed.")
        return errs


default_app_config = 'pretix_bitpay.BitpayApp'
